from datetime import timedelta

from django.utils import timezone

from scheduler.models.task import TaskType
from scheduler.tests.testtools import SchedulerBaseCase, task_factory
from scheduler.tools import get_scheduled_task


class TestInternals(SchedulerBaseCase):
    def test_get_scheduled_job(self):
        task = task_factory(TaskType.ONCE, scheduled_time=timezone.now() - timedelta(hours=1))
        self.assertEqual(task, get_scheduled_task(TaskType.ONCE, task.id))
        with self.assertRaises(ValueError):
            get_scheduled_task(task.task_type, task.id + 1)
        with self.assertRaises(ValueError):
            get_scheduled_task("UNKNOWN_JOBTYPE", task.id)



from django.core.management import call_command
from django.test import TestCase

from scheduler.tests import test_settings # noqa


class RqstatsTest(TestCase):
    def test_rqstats__does_not_fail(self):
        call_command("rqstats", "-j")
        call_command("rqstats", "-y")
        call_command("rqstats")



from django.core.management import call_command

from scheduler.queues import get_queue
from scheduler.tests.jobs import failing_job
from scheduler.tests.test_views import BaseTestCase
from scheduler.tools import create_worker
from scheduler.tests import test_settings # noqa


class DeleteFailedExecutionsTest(BaseTestCase):
    def test_delete_failed_executions__delete_jobs(self):
        queue = get_queue("default")
        call_command("delete_failed_executions", queue="default")
        queue.enqueue(failing_job)
        worker = create_worker("default")
        worker.work(burst=True)
        self.assertEqual(1, len(queue.failed_job_registry))
        call_command("delete_failed_executions", queue="default")
        self.assertEqual(0, len(queue.failed_job_registry))



from .ephemeral_models import QueueAdmin, WorkerAdmin
from .old_task_models import TaskAdmin as OldTaskAdmin
from .task_admin import TaskAdmin

__all__ = ["OldTaskAdmin", "QueueAdmin", "WorkerAdmin", "TaskAdmin", ]



# Generated by Django 5.1.3 on 2024-11-20 20:32

import django.db.models.deletion
import scheduler.models.task
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("scheduler", "0018_alter_crontask_queue_alter_repeatabletask_queue_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="Task",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                (
                    "name",
                    models.CharField(help_text="Name of the job", max_length=128, unique=True, verbose_name="name"),
                ),
                (
                    "task_type",
                    models.CharField(
                        choices=[
                            ("CronTaskType", "Cron Task"),
                            ("RepeatableTaskType", "Repeatable Task"),
                            ("OnceTaskType", "Run once"),
                        ],
                        default="OnceTaskType",
                        max_length=32,
                        verbose_name="Task type",
                    ),
                ),
                ("callable", models.CharField(max_length=2048, verbose_name="callable")),
                (
                    "enabled",
                    models.BooleanField(
                        default=True,
                        help_text="Should job be scheduled? This field is useful to keep past jobs that should no longer be scheduled",
                        verbose_name="enabled",
                    ),
                ),
                (
                    "queue",
                    models.CharField(
                        choices=scheduler.models.task.get_queue_choices,
                        help_text="Queue name",
                        max_length=255,
                        verbose_name="queue",
                    ),
                ),
                (
                    "job_id",
                    models.CharField(
                        blank=True,
                        editable=False,
                        help_text="Current job_id on queue",
                        max_length=128,
                        null=True,
                        verbose_name="job id",
                    ),
                ),
                (
                    "at_front",
                    models.BooleanField(
                        default=False,
                        help_text="When queuing the job, add it in the front of the queue",
                        verbose_name="At front",
                    ),
                ),
                (
                    "timeout",
                    models.IntegerField(
                        blank=True,
                        help_text="Timeout specifies the maximum runtime, in seconds, for the job before it'll be considered 'lost'. Blank uses the default timeout.",
                        null=True,
                        verbose_name="timeout",
                    ),
                ),
                (
                    "result_ttl",
                    models.IntegerField(
                        blank=True,
                        help_text="The TTL value (in seconds) of the job result.<br/>\n               -1: Result never expires, you should delete jobs manually. <br/>\n                0: Result gets deleted immediately. <br/>\n                >0: Result expires after n seconds.",
                        null=True,
                        verbose_name="result ttl",
                    ),
                ),
                (
                    "failed_runs",
                    models.PositiveIntegerField(
                        default=0, help_text="Number of times the task has failed", verbose_name="failed runs"
                    ),
                ),
                (
                    "successful_runs",
                    models.PositiveIntegerField(
                        default=0, help_text="Number of times the task has succeeded", verbose_name="successful runs"
                    ),
                ),
                (
                    "last_successful_run",
                    models.DateTimeField(
                        blank=True,
                        help_text="Last time the task has succeeded",
                        null=True,
                        verbose_name="last successful run",
                    ),
                ),
                (
                    "last_failed_run",
                    models.DateTimeField(
                        blank=True, help_text="Last time the task has failed", null=True, verbose_name="last failed run"
                    ),
                ),
                (
                    "interval",
                    models.PositiveIntegerField(
                        blank=True, help_text="Interval for repeatable task", null=True, verbose_name="interval"
                    ),
                ),
                (
                    "interval_unit",
                    models.CharField(
                        blank=True,
                        choices=[
                            ("seconds", "seconds"),
                            ("minutes", "minutes"),
                            ("hours", "hours"),
                            ("days", "days"),
                            ("weeks", "weeks"),
                        ],
                        default="hours",
                        max_length=12,
                        null=True,
                        verbose_name="interval unit",
                    ),
                ),
                (
                    "repeat",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Number of times to run the job. Leaving this blank means it will run forever.",
                        null=True,
                        verbose_name="repeat",
                    ),
                ),
                ("scheduled_time", models.DateTimeField(blank=True, null=True, verbose_name="scheduled time")),
                (
                    "cron_string",
                    models.CharField(
                        blank=True,
                        help_text='Define the schedule in a crontab like syntax.\n            Times are in UTC. Use <a href="https://crontab.guru/">crontab.guru</a> to create a cron string.',
                        max_length=64,
                        null=True,
                        verbose_name="cron string",
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="crontask",
            name="new_task_id",
            field=models.ForeignKey(
                blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to="scheduler.task"
            ),
        ),
        migrations.AddField(
            model_name="repeatabletask",
            name="new_task_id",
            field=models.ForeignKey(
                blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to="scheduler.task"
            ),
        ),
        migrations.AddField(
            model_name="scheduledtask",
            name="new_task_id",
            field=models.ForeignKey(
                blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to="scheduler.task"
            ),
        ),
    ]



from datetime import datetime
from typing import Dict, Any, Optional

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.utils import timezone

from scheduler.models.old_scheduled_task import BaseTask
from scheduler.models.task import Task, TaskArg, TaskKwarg
from scheduler.settings import logger
from scheduler.tools import TaskType


def job_model_str(model_str: str) -> str:
    if model_str.find("Job") == len(model_str) - 3:
        return model_str[:-3] + "Task"
    return model_str


def get_task_type(model_str: str) -> TaskType:
    model_str = job_model_str(model_str)
    try:
        return TaskType(model_str)
    except ValueError:
        pass
    if model_str == "CronTask":
        return TaskType.CRON
    elif model_str == "RepeatableTask":
        return TaskType.REPEATABLE
    elif model_str in {"ScheduledTask", "OnceTask"}:
        return TaskType.ONCE
    raise ValueError(f"Invalid model {model_str}")


def create_task_from_dict(task_dict: Dict[str, Any], recreate: bool) -> Optional[Task]:
    existing_task = None
    if "new_task_id" in task_dict:
        existing_task = Task.objects.filter(id=task_dict["new_task_id"]).first()
    if existing_task is None:
        existing_task = Task.objects.filter(name=task_dict["name"]).first()
    task_type = get_task_type(task_dict["model"])
    if existing_task:
        if recreate:
            logger.info(f'Found existing job "{existing_task}, removing it to be reinserted"')
            existing_task.delete()
        else:
            logger.info(f'Found existing job "{existing_task}", skipping')
            return existing_task
    kwargs = dict(task_dict)
    kwargs["task_type"] = task_type
    del kwargs["model"]
    del kwargs["callable_args"]
    del kwargs["callable_kwargs"]
    del kwargs["new_task_id"]
    if kwargs.get("scheduled_time", None):
        target = datetime.fromisoformat(kwargs["scheduled_time"])
        if not settings.USE_TZ and not timezone.is_naive(target):
            target = timezone.make_naive(target)
        kwargs["scheduled_time"] = target
    model_fields = filter(lambda field: hasattr(field, 'attname'), Task._meta.get_fields())
    model_fields = set(map(lambda field: field.attname, model_fields))
    keys_to_ignore = list(filter(lambda _k: _k not in model_fields, kwargs.keys()))
    for k in keys_to_ignore:
        del kwargs[k]
    task = Task.objects.create(**kwargs)
    logger.info(f"Created task {task}")
    content_type = ContentType.objects.get_for_model(task)

    for arg in task_dict["callable_args"]:
        TaskArg.objects.create(
            content_type=content_type,
            object_id=task.id,
            **arg,
        )
    for arg in task_dict["callable_kwargs"]:
        TaskKwarg.objects.create(
            content_type=content_type,
            object_id=task.id,
            **arg,
        )
    return task


def migrate(old: BaseTask) -> Optional[Task]:
    old_task_dict = old.to_dict()
    new_task = create_task_from_dict(old_task_dict, old_task_dict.get("new_task_id") is not None)
    old.new_task_id = new_task
    old.enabled = False
    old.save()
    return new_task



from django.core.management import call_command
from django.test import TestCase

from scheduler.queues import get_queue
from scheduler.tests.jobs import test_job
from scheduler.tests import test_settings # noqa


class RunJobTest(TestCase):
    def test_run_job__should_schedule_job(self):
        queue = get_queue("default")
        queue.empty()
        func_name = f"{test_job.__module__}.{test_job.__name__}"
        # act
        call_command("run_job", func_name, queue="default")
        # assert
        job_list = queue.get_jobs()
        self.assertEqual(1, len(job_list))
        self.assertEqual(func_name + "()", job_list[0].get_call_string())



from typing import List, Optional, Union

import django
from django.apps import apps
from rq import Worker
from rq.command import send_stop_job_command
from rq.decorators import job
from rq.exceptions import InvalidJobOperation
from rq.job import Job, JobStatus
from rq.job import get_current_job  # noqa
from rq.queue import Queue, logger
from rq.registry import (
    DeferredJobRegistry,
    FailedJobRegistry,
    FinishedJobRegistry,
    ScheduledJobRegistry,
    StartedJobRegistry,
    CanceledJobRegistry,
    BaseRegistry,
)
from rq.scheduler import RQScheduler
from rq.worker import WorkerStatus

from scheduler import settings
from scheduler.broker_types import PipelineType, ConnectionType

MODEL_NAMES = ["ScheduledTask", "RepeatableTask", "CronTask", "Task"]
TASK_TYPES = ["OnceTaskType", "RepeatableTaskType", "CronTaskType"]

rq_job_decorator = job
ExecutionStatus = JobStatus
InvalidJobOperation = InvalidJobOperation


def register_sentry(sentry_dsn, **opts):
    from rq.contrib.sentry import register_sentry as rq_register_sentry

    rq_register_sentry(sentry_dsn, **opts)


def as_str(v: Union[bytes, str]) -> Optional[str]:
    """Converts a `bytes` value to a string using `utf-8`.

    :param v: The value (None/bytes/str)
    :raises: ValueError: If the value is not `bytes` or `str`
    :returns: Either the decoded string or None
    """
    if v is None:
        return None
    if isinstance(v, bytes):
        return v.decode("utf-8")
    if isinstance(v, str):
        return v
    raise ValueError("Unknown type %r" % type(v))


class JobExecution(Job):
    def __eq__(self, other) -> bool:
        return isinstance(other, Job) and self.id == other.id

    @property
    def is_scheduled_task(self) -> bool:
        return self.meta.get("scheduled_task_id", None) is not None

    def is_execution_of(self, task: "Task") -> bool:  # noqa: F821
        return (
                self.meta.get("task_type", None) == task.task_type and self.meta.get("scheduled_task_id",
                                                                                     None) == task.id
        )

    def stop_execution(self, connection: ConnectionType):
        send_stop_job_command(connection, self.id)


class DjangoWorker(Worker):
    def __init__(self, *args, **kwargs):
        self.fork_job_execution = kwargs.pop("fork_job_execution", True)
        job_class = kwargs.get("job_class") or JobExecution
        if not isinstance(job_class, type) or not issubclass(job_class, JobExecution):
            raise ValueError("job_class must be a subclass of JobExecution")

        # Update kwargs with the potentially modified job_class
        kwargs["job_class"] = job_class
        kwargs["queue_class"] = DjangoQueue
        super(DjangoWorker, self).__init__(*args, **kwargs)

    def __eq__(self, other):
        return isinstance(other, Worker) and self.key == other.key and self.name == other.name

    def __hash__(self):
        return hash((self.name, self.key, ",".join(self.queue_names())))

    def __str__(self):
        return f"{self.name}/{','.join(self.queue_names())}"

    def _start_scheduler(
            self,
            burst: bool = False,
            logging_level: str = "INFO",
            date_format: str = "%H:%M:%S",
            log_format: str = "%(asctime)s %(message)s",
    ) -> None:
        """Starts the scheduler process.
        This is specifically designed to be run by the worker when running the `work()` method.
        Instantiates the DjangoScheduler and tries to acquire a lock.
        If the lock is acquired, start scheduler.
        If the worker is on burst mode, just enqueues scheduled jobs and quits,
        otherwise, starts the scheduler in a separate process.

        :param burst (bool, optional): Whether to work on burst mode. Defaults to False.
        :param logging_level (str, optional): Logging level to use. Defaults to "INFO".
        :param date_format (str, optional): Date Format. Defaults to DEFAULT_LOGGING_DATE_FORMAT.
        :param log_format (str, optional): Log Format. Defaults to DEFAULT_LOGGING_FORMAT.
        """
        self.scheduler = DjangoScheduler(
            self.queues,
            connection=self.connection,
            logging_level=logging_level,
            date_format=date_format,
            log_format=log_format,
            serializer=self.serializer,
        )
        self.scheduler.acquire_locks()
        if self.scheduler.acquired_locks:
            if burst:
                self.scheduler.enqueue_scheduled_jobs()
                self.scheduler.release_locks()
            else:
                proc = self.scheduler.start()
                self._set_property("scheduler_pid", proc.pid)

    def execute_job(self, job: "Job", queue: "Queue") -> None:
        if self.fork_job_execution:
            super(DjangoWorker, self).execute_job(job, queue)
        else:
            self.set_state(WorkerStatus.BUSY)
            self.perform_job(job, queue)
            self.set_state(WorkerStatus.IDLE)

    def work(self, **kwargs) -> bool:
        kwargs.setdefault("with_scheduler", True)
        return super(DjangoWorker, self).work(**kwargs)

    def _set_property(self, prop_name: str, val, pipeline: Optional[PipelineType] = None) -> None:
        connection = pipeline if pipeline is not None else self.connection
        if val is None:
            connection.hdel(self.key, prop_name)
        else:
            connection.hset(self.key, prop_name, val)

    def _get_property(self, prop_name: str, pipeline: Optional[PipelineType] = None) -> Optional[str]:
        connection = pipeline if pipeline is not None else self.connection
        res = connection.hget(self.key, prop_name)
        return as_str(res)

    def scheduler_pid(self) -> Optional[int]:
        if len(self.queues) == 0:
            logger.warning("No queues to get scheduler pid from")
            return None
        pid = self.connection.get(DjangoScheduler.get_locking_key(self.queues[0].name))
        return int(pid.decode()) if pid is not None else None


class DjangoQueue(Queue):
    """A subclass of RQ's QUEUE that allows jobs to be stored temporarily to be enqueued later at the end of Django's
    request/response cycle."""

    REGISTRIES = dict(
        finished="finished_job_registry",
        failed="failed_job_registry",
        scheduled="scheduled_job_registry",
        started="started_job_registry",
        deferred="deferred_job_registry",
        canceled="canceled_job_registry",
    )

    def __init__(self, *args, **kwargs) -> None:
        kwargs["job_class"] = JobExecution
        super(DjangoQueue, self).__init__(*args, **kwargs)

    def get_registry(self, name: str) -> Union[None, BaseRegistry, "DjangoQueue"]:
        name = name.lower()
        if name == "queued":
            return self
        elif name in DjangoQueue.REGISTRIES:
            return getattr(self, DjangoQueue.REGISTRIES[name])
        return None

    @property
    def finished_job_registry(self) -> FinishedJobRegistry:
        return FinishedJobRegistry(self.name, self.connection)

    @property
    def started_job_registry(self) -> StartedJobRegistry:
        return StartedJobRegistry(
            self.name,
            self.connection,
            job_class=JobExecution,
        )

    @property
    def deferred_job_registry(self) -> DeferredJobRegistry:
        return DeferredJobRegistry(
            self.name,
            self.connection,
            job_class=JobExecution,
        )

    @property
    def failed_job_registry(self) -> FailedJobRegistry:
        return FailedJobRegistry(
            self.name,
            self.connection,
            job_class=JobExecution,
        )

    @property
    def scheduled_job_registry(self) -> ScheduledJobRegistry:
        return ScheduledJobRegistry(
            self.name,
            self.connection,
            job_class=JobExecution,
        )

    @property
    def canceled_job_registry(self) -> CanceledJobRegistry:
        return CanceledJobRegistry(
            self.name,
            self.connection,
            job_class=JobExecution,
        )

    def get_all_job_ids(self) -> List[str]:
        res = list()
        res.extend(self.get_job_ids())
        res.extend(self.finished_job_registry.get_job_ids())
        res.extend(self.started_job_registry.get_job_ids())
        res.extend(self.deferred_job_registry.get_job_ids())
        res.extend(self.failed_job_registry.get_job_ids())
        res.extend(self.scheduled_job_registry.get_job_ids())
        res.extend(self.canceled_job_registry.get_job_ids())
        return res

    def get_all_jobs(self) -> List[JobExecution]:
        job_ids = self.get_all_job_ids()
        return list(filter(lambda j: j is not None, [self.fetch_job(job_id) for job_id in job_ids]))

    def clean_registries(self) -> None:
        self.started_job_registry.cleanup()
        self.failed_job_registry.cleanup()
        self.finished_job_registry.cleanup()

    def remove_job_id(self, job_id: str) -> None:
        self.connection.lrem(self.key, 0, job_id)

    def last_job_id(self) -> Optional[str]:
        return self.connection.lindex(self.key, 0)


class DjangoScheduler(RQScheduler):
    def __init__(self, *args, **kwargs) -> None:
        kwargs.setdefault("interval", settings.SCHEDULER_CONFIG.SCHEDULER_INTERVAL)
        super(DjangoScheduler, self).__init__(*args, **kwargs)

    @staticmethod
    def reschedule_all_jobs():
        for model_name in MODEL_NAMES:
            model = apps.get_model(app_label="scheduler", model_name=model_name)
            enabled_jobs = model.objects.filter(enabled=True)
            for item in enabled_jobs:
                logger.debug(f"Rescheduling {str(item)}")
                item.save()

    def work(self) -> None:
        django.setup()
        super(DjangoScheduler, self).work()

    def enqueue_scheduled_jobs(self) -> None:
        self.reschedule_all_jobs()
        super(DjangoScheduler, self).enqueue_scheduled_jobs()



from django.core.management import call_command
from django.test import TestCase

from scheduler.queues import get_queue
from scheduler.tests.jobs import failing_job
from scheduler.tests import test_settings # noqa


class RqworkerTestCase(TestCase):

    def test_rqworker__no_queues_params(self):
        queue = get_queue("default")

        # enqueue some jobs that will fail
        jobs = []
        job_ids = []
        for _ in range(0, 3):
            job = queue.enqueue(failing_job)
            jobs.append(job)
            job_ids.append(job.id)

        # Create a worker to execute these jobs
        call_command("rqworker", fork_job_execution=False, burst=True)

        # check if all jobs are really failed
        for job in jobs:
            self.assertTrue(job.is_failed)

    def test_rqworker__job_class_param__green(self):
        queue = get_queue("default")

        # enqueue some jobs that will fail
        jobs = []
        job_ids = []
        for _ in range(0, 3):
            job = queue.enqueue(failing_job)
            jobs.append(job)
            job_ids.append(job.id)

        # Create a worker to execute these jobs
        call_command(
            "rqworker", "--job-class", "scheduler.rq_classes.JobExecution", fork_job_execution=False, burst=True
        )

        # check if all jobs are really failed
        for job in jobs:
            self.assertTrue(job.is_failed)

    def test_rqworker__bad_job_class__fail(self):
        queue = get_queue("default")

        # enqueue some jobs that will fail
        jobs = []
        job_ids = []
        for _ in range(0, 3):
            job = queue.enqueue(failing_job)
            jobs.append(job)
            job_ids.append(job.id)

        # Create a worker to execute these jobs
        with self.assertRaises(ImportError):
            call_command("rqworker", "--job-class", "rq.badclass", fork_job_execution=False, burst=True)

    def test_rqworker__run_jobs(self):
        queue = get_queue("default")

        # enqueue some jobs that will fail
        jobs = []
        job_ids = []
        for _ in range(0, 3):
            job = queue.enqueue(failing_job)
            jobs.append(job)
            job_ids.append(job.id)

        # Create a worker to execute these jobs
        call_command("rqworker", "default", fork_job_execution=False, burst=True)

        # check if all jobs are really failed
        for job in jobs:
            self.assertTrue(job.is_failed)

    def test_rqworker__worker_with_two_queues(self):
        queue = get_queue("default")
        queue2 = get_queue("django_tasks_scheduler_test")

        # enqueue some jobs that will fail
        jobs = []
        job_ids = []
        for _ in range(0, 3):
            job = queue.enqueue(failing_job)
            jobs.append(job)
            job_ids.append(job.id)
        job = queue2.enqueue(failing_job)
        jobs.append(job)
        job_ids.append(job.id)

        # Create a worker to execute these jobs
        call_command("rqworker", "default", "django_tasks_scheduler_test", fork_job_execution=False, burst=True)

        # check if all jobs are really failed
        for job in jobs:
            self.assertTrue(job.is_failed)

    def test_rqworker__worker_with_one_queue__does_not_perform_other_queue_job(self):
        queue = get_queue("default")
        queue2 = get_queue("django_tasks_scheduler_test")

        job = queue.enqueue(failing_job)
        other_job = queue2.enqueue(failing_job)

        # Create a worker to execute these jobs
        call_command("rqworker", "default", fork_job_execution=False, burst=True)
        # assert
        self.assertTrue(job.is_failed)
        self.assertTrue(other_job.is_queued)



# This is a helper module to obfuscate types used by different broker implementations.
from typing import Union, Dict, Tuple, Type

import redis

try:
    import valkey
except ImportError:
    valkey = redis
    valkey.Valkey = redis.Redis
    valkey.StrictValkey = redis.StrictRedis

from scheduler.settings import Broker

ConnectionErrorTypes = (redis.ConnectionError, valkey.ConnectionError)
ResponseErrorTypes = (redis.ResponseError, valkey.ResponseError)
ConnectionType = Union[redis.Redis, valkey.Valkey]
PipelineType = Union[redis.client.Pipeline, valkey.client.Pipeline]
SentinelType = Union[redis.sentinel.Sentinel, valkey.sentinel.Sentinel]

BrokerMetaData: Dict[Tuple[Broker, bool], Tuple[Type[ConnectionType], Type[SentinelType], str]] = {
    # Map of (Broker, Strict flag) => Connection Class, Sentinel Class, SSL Connection Prefix
    (Broker.REDIS, False): (redis.Redis, redis.sentinel.Sentinel, "rediss"),
    (Broker.VALKEY, False): (valkey.Valkey, valkey.sentinel.Sentinel, "valkeys"),
    (Broker.REDIS, True): (redis.StrictRedis, redis.sentinel.Sentinel, "rediss"),
    (Broker.VALKEY, True): (valkey.StrictValkey, valkey.sentinel.Sentinel, "valkeys"),
}



import importlib
import os
from typing import List, Any, Callable, Optional

import croniter
from django.apps import apps
from django.db import models
from django.utils import timezone
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _

from scheduler.queues import get_queues, logger, get_queue
from scheduler.rq_classes import DjangoWorker, JobExecution, TASK_TYPES, MODEL_NAMES
from scheduler.settings import SCHEDULER_CONFIG, Broker


class TaskType(models.TextChoices):
    CRON = "CronTaskType", _("Cron Task")
    REPEATABLE = "RepeatableTaskType", _("Repeatable Task")
    ONCE = "OnceTaskType", _("Run once")


def callable_func(callable_str: str) -> Callable:
    path = callable_str.split(".")
    module = importlib.import_module(".".join(path[:-1]))
    func = getattr(module, path[-1])
    if callable(func) is False:
        raise TypeError("'{}' is not callable".format(callable_str))
    return func


def get_next_cron_time(cron_string: Optional[str]) -> Optional[timezone.datetime]:
    """Calculate the next scheduled time by creating a crontab object with a cron string"""
    if cron_string is None:
        return None
    now = timezone.now()
    itr = croniter.croniter(cron_string, now)
    next_itr = itr.get_next(timezone.datetime)
    return next_itr


def get_scheduled_task(task_type_str: str, task_id: int) -> "BaseTask":  # noqa: F821
    # Try with new model names
    model = apps.get_model(app_label="scheduler", model_name="Task")
    if task_type_str in TASK_TYPES:
        try:
            task_type = TaskType(task_type_str)
            task = model.objects.filter(task_type=task_type, id=task_id).first()
            if task is None:
                raise ValueError(f"Job {task_type}:{task_id} does not exit")
            return task
        except ValueError:
            raise ValueError(f"Invalid task type {task_type_str}")
    elif task_type_str in MODEL_NAMES:
        model = apps.get_model(app_label="scheduler", model_name=task_type_str)
        task = model.objects.filter(id=task_id).first()
        if task is None:
            raise ValueError(f"Job {task_type_str}:{task_id} does not exit")
        return task
    raise ValueError(f"Job Model {task_type_str} does not exist, choices are {TASK_TYPES}")


def run_task(task_model: str, task_id: int) -> Any:
    """Run a scheduled job"""
    scheduled_task = get_scheduled_task(task_model, task_id)
    logger.debug(f"Running task {str(scheduled_task)}")
    args = scheduled_task.parse_args()
    kwargs = scheduled_task.parse_kwargs()
    res = scheduled_task.callable_func()(*args, **kwargs)
    return res


def _calc_worker_name(existing_worker_names) -> str:
    hostname = os.uname()[1]
    c = 1
    worker_name = f"{hostname}-worker.{c}"
    while worker_name in existing_worker_names:
        c += 1
        worker_name = f"{hostname}-worker.{c}"
    return worker_name


def create_worker(*queue_names, **kwargs) -> DjangoWorker:
    """Returns a Django worker for all queues or specified ones."""

    queues = get_queues(*queue_names)
    existing_workers = DjangoWorker.all(connection=queues[0].connection)
    existing_worker_names = set(map(lambda w: w.name, existing_workers))
    kwargs.setdefault("fork_job_execution", SCHEDULER_CONFIG.BROKER != Broker.FAKEREDIS)
    if kwargs.get("name", None) is None:
        kwargs["name"] = _calc_worker_name(existing_worker_names)

    kwargs["name"] = kwargs["name"].replace("/", ".")

    # Handle job_class if provided
    if "job_class" not in kwargs or kwargs["job_class"] is None:
        kwargs["job_class"] = "scheduler.rq_classes.JobExecution"
    try:
        kwargs["job_class"] = import_string(kwargs["job_class"])
    except ImportError:
        raise ImportError(f"Could not import job class {kwargs['job_class']}")

    worker = DjangoWorker(queues, connection=queues[0].connection, **kwargs)
    return worker


def get_job_executions_for_task(queue_name, scheduled_task) -> List[JobExecution]:
    queue = get_queue(queue_name)
    job_list = queue.get_all_jobs()
    res = list(filter(lambda j: j.is_execution_of(scheduled_task), job_list))
    return res



from datetime import timedelta

from django.utils import timezone

from scheduler import settings
from scheduler.models.task import TaskType
from scheduler.tests.test_task_types.test_task_model import BaseTestCases
from scheduler.tests.testtools import task_factory


class TestScheduledTask(BaseTestCases.TestSchedulableTask):
    task_type = TaskType.ONCE

    def test_clean(self):
        job = task_factory(self.task_type)
        job.queue = list(settings.QUEUES)[0]
        job.callable = "scheduler.tests.jobs.test_job"
        self.assertIsNone(job.clean())

    def test_unschedulable_old_job(self):
        job = task_factory(self.task_type, scheduled_time=timezone.now() - timedelta(hours=1))
        self.assertFalse(job.is_scheduled())
